using System;
using NUnit.Framework;
using System.Diagnostics;
using System.Collections;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using JMBucknall.Containers;

namespace Tests {
    [Serializable()]
    public class ReversePriority : IComparable {
        private int priority;

        public ReversePriority(int priority) {
            this.priority = priority;
        }

        public int CompareTo(object obj) {
            ReversePriority castObj = obj as ReversePriority;
            if (castObj == null) 
                throw new InvalidCastException();
            if (priority < castObj.priority) 
                return 1;
            if (priority == castObj.priority)
                return 0;
            return -1;
        }
    }



    [TestFixture] 
    public class TestPQ {
        PriorityQueue pq;

        [SetUp] public void SetUp() {
            pq = new PriorityQueue();
        }

        [TearDown] public void TearDown() {
        }

        [Test] public void TestEnqueueing() {
            pq.Enqueue("item one", 12);
            Assertion.AssertEquals(
                "Enqueuing first item should set count to 1",
                1, pq.Count);
            pq.Enqueue("item two", 5);
            Assertion.AssertEquals(
                "Enqueuing second item should set count to 2",
                2, pq.Count);
        }

        [Test] public void TestDequeueing() {
            pq.Enqueue("item one", 12);
            pq.Enqueue("item two", 5);
            string s = (string) pq.Dequeue();
            Assertion.AssertEquals(
                "Dequeuing item should set count to 1",
                1, pq.Count);
            Assertion.AssertEquals(
                "Dequeuing should retrieve highest priority item",
                "item one", s);
            s = (string) pq.Dequeue();
            Assertion.AssertEquals(
                "Dequeuing final item should set count to 0",
                0, pq.Count);
            Assertion.AssertEquals(
                "Dequeuing should retrieve next item",
                "item two", s);
        }

        [ExpectedException(typeof(InvalidOperationException))]
        [Test] public void TestDequeueWithEmptyQueue() {
            pq.Dequeue();
        }

        [Test] public void TestGrowingQueue() {
            string s;
            string expectedStr;
            for (int i = 0; i < 15; i++) {
                pq.Enqueue("item: " + i.ToString(), i * 2);
            }
            Assertion.AssertEquals("Enqueued 15 items, so there should be 15 there", 15, pq.Count);
            pq.Enqueue("trigger", 15);
            Assertion.AssertEquals("Enqueued another, so there should be 16 there", 16, pq.Count);
            for (int i = 14; i > 7; i--) {
                s = (string) pq.Dequeue();
                expectedStr = "item: " + i.ToString();
                Assertion.AssertEquals("Dequeueing problem", expectedStr, s);
            }
            s = (string) pq.Dequeue();
            Assertion.AssertEquals("Dequeueing problem", "trigger", s);
            for (int i = 7; i >= 0; i--) {
                s = (string) pq.Dequeue();
                expectedStr = "item: " + i.ToString();
                Assertion.AssertEquals("Dequeing problem", expectedStr, s);
            }
        }

        [Test] public void TestEnumerator() {
            string s;
            // use a hashtable to check contents of PQ
            Hashtable ht = new Hashtable(); 
            for (int i = 0; i < 5; i++) {
                s = "item: " + i.ToString();
                ht.Add(s, null);
                pq.Enqueue(s, i * 2);
            }
            foreach(string str in pq) {
                Assertion.AssertEquals("String from PQ wasn't found in test hashtable",
                    true, ht.Contains(str));
                ht.Remove(str); // so we don't see it again
            }
            Assertion.AssertEquals("the hashtable should be empty at the end",
                0, ht.Count);
        }

        [ExpectedException(typeof(InvalidOperationException))]
        [Test] public void TestEnumeratorWithEnqueue() {
            pq.Enqueue("one", 42);
            IEnumerator ie = pq.GetEnumerator();
            ie.MoveNext();
            pq.Enqueue("two", 13);
            ie.MoveNext(); // should fail
            Assertion.Fail("The previous call to MoveNext() should fail");
        }

        [ExpectedException(typeof(InvalidOperationException))]
        [Test] public void TestEnumeratorWithDequeue() {
            pq.Enqueue("one", 42);
            IEnumerator ie = pq.GetEnumerator();
            ie.MoveNext();
            pq.Dequeue();
            ie.MoveNext(); // should fail
            Assertion.Fail("The previous call to MoveNext() should fail");
        }

        [Test] public void TestCopyTo() {
            string s;
            // use a hashtable to check contents of returned array
            Hashtable ht = new Hashtable(); 
            for (int i = 0; i < 5; i++) {
                s = "item: " + i.ToString();
                ht.Add(s, null);
                pq.Enqueue(s, i * 2);
            }

            HeapEntry[] heArray = new HeapEntry[6];

            pq.CopyTo(heArray, 1);

            foreach(HeapEntry he in heArray) {
                if (he.Item != null) {
                    Assertion.AssertEquals("String from PQ wasn't found in test hashtable",
                        true, ht.Contains(he.Item));
                    ht.Remove(he.Item); // so we don't see it again
                }
            }
            Assertion.AssertEquals("the hashtable should be empty at the end",
                0, ht.Count);
        }

        [Test] public void TestPriorityType() {
            pq.Enqueue("item one", new ReversePriority(12));
            pq.Enqueue("item two", new ReversePriority(5));
            string s = (string) pq.Dequeue();
            Assertion.AssertEquals(
                "Dequeuing item should set count to 1",
                1, pq.Count);
            Assertion.AssertEquals(
                "Dequeuing should retrieve highest priority item",
                "item two", s);
            s = (string) pq.Dequeue();
            Assertion.AssertEquals(
                "Dequeuing final item should set count to 0",
                0, pq.Count);
            Assertion.AssertEquals(
                "Dequeuing should retrieve next item",
                "item one", s);
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [Test] public void TestEnqueueWithNullPriority() {
            pq.Enqueue("A", null);
        }

        [Test] public void TestSerialization() {
            pq.Enqueue("item one", new ReversePriority(12));
            pq.Enqueue("item two", new ReversePriority(5));


            Stream s = new MemoryStream();
            IFormatter f = new BinaryFormatter();

            f.Serialize(s, pq);
            s.Position = 0;
            PriorityQueue pq2 = (PriorityQueue) f.Deserialize(s);
            s.Close();

            string str = (string) pq2.Dequeue();
            Assertion.AssertEquals(
                "Dequeuing item should set count to 1",
                1, pq2.Count);
            Assertion.AssertEquals(
                "Dequeuing should retrieve highest priority item",
                "item two", str);
            str = (string) pq2.Dequeue();
            Assertion.AssertEquals(
                "Dequeuing final item should set count to 0",
                0, pq2.Count);
            Assertion.AssertEquals(
                "Dequeuing should retrieve next item",
                "item one", str);
        }
    }
}